/********************************************************************
 *******************************************************************/


/********************************************************************
To use this demo, program a device with the firmware from this project.
Then plug the device into a Windows 7 or Windows Vista SP1+ based PC.

MSDN documentation for the "WM_GESTURE" messages indicate that they are
currently only supported in Windows 7.  WM_GESTURE messages can be used
by PC applications to receive information related to a user's input.  
However, Windows Vista SP1 still has some digitizer input 
capabilities, and can therefore still be used to experiment with this demo.
The demo experience will be superior under Windows 7 however, as not
all touch related capabilities and features are enabled in Windows Vista.

The device should enumerate as a HID class multi-touch capable
digitizer device.  Windows 7 and Windows Vista features associated 
with "Windows Touch" should automatically become activated.  
Under default versions of Windows XP, the device should still enumerate 
as a HID class device, but no new digitizer input functionality will 
be activated.

Digitizer tablets send absolute contact position information
to the host.  This is a bit different compared to conventional mouse
devices, in that mice send relative X/Y direction movement information,
instead of an absolute location.

Since the demo boards that this firmware is intended to be run on
do not include an actual multi-touch tablet area, the firmware simply
emulates the tablet by sending the same data (and formatting) that a tablet
would actually be used to produce.  The demo board pushbutton "sw3"
(see definition in the HardwareProfile - (board name).h file)
can be used to make the demo board send simulated gesture info to the host.

To use this demo, open up an Internet Explorer window, and go to some website 
(ex: www.microchip.com/usb).
The web page opened should be large enough that both vertical and 
horizontal scrollbars will become active when zoomed in. Click on some 
link, and then click the back button.  The purpose of this is to enable both 
the forward and back buttons in the web broswer.  
This demo will send gesture information to the OS that can be interpreted as 
forward and back "flick" gestures, which can be used to control the web browser.

Now try pressing the pushbutton on the demo board repeatedly.  Each press
of the pushbutton should send a bunch of IN packets to the host, which contain
the information that would typically be generated by a human user interacting 
with a multi-touch capable digitizer screen.

The firmware will send traffic to emulate both single touch and multi-touch 
(up to 2 simultaneous contacts) gestures, based on the number of times the
pushbutton has been pressed.  Under Windows 7 (Vista will be somewhat different),
the following gestures will be sent and should be recognized by Internet Explorer:

1.  Emulate a single touch "back" gesture operation.  
		This is done by pressing only one finger down, and by moving it right quickly (X plus)
2.  Emulate a single touch "forward" gesture operation.  
		This is done by pressing only one finger down, and by moving it left quickly (X minus)
3.  Emulate a multi-touch zoom-in operation ("un-pinch" index and thumb fingers)
4.  Emulate a multi-touch zoom-in operation ("un-pinch" index and thumb fingers)
5.  Emulate a multi-touch zoom-in operation ("un-pinch" index and thumb fingers)
6.  Emulate a multi-touch horizontal scroll right operation 
      This is done by using two fingers simultaneously and moving them both to the left (X minus)
7.  Emulate a multi-touch horizontal scroll left operation
      This is done by using two fingers simultaneously and moving them both to the right (X plus)
8.  Emulate a single touch scroll down operation.  
      This is done by pressing only one finger down, and by moving it down (Y plus)
9.  Emulate a single touch scroll up operation.  
      This is done by pressing only one finger down, and by moving it up (Y minus)
10. Emulate a multi-touch zoom-out operation (pinch two fingers together)

11+. Upon pressing the button 11 times, the firmware will go back to state 1. Subsequent 
       presses of the button will step through and repeat the above gestures 1-10.

Note: The above demo zooms in three times in a row (presses 3-5).  The purpose of this
is to zoom in sufficiently so that the horizontal and vertical scoll bars will become
active.  The gestures associated with button presses 6-9 perform scroll operations,
but only if the scroll bars are active.

Note2: This demo was tested and developed primarily on 64-bit Windows 7 beta build 7000, 
with some additional testing on Windows Vista SP1.  At the time of this development,
Windows 7 is not yet a finalized and released product.  If you have trouble getting
this demo to work correctly, check to see if there is a later version of this demo
available from the Microchip USB design center: www.microchip.com/usb

Additionally, try adjusting some of the user adjustable controls for 
Pen and Touch Input Devices.  These adjustments can be found from the control panel 
within Windows.


This demo is configured by default to support up to two simultaneous contacts,
by using "parallel reporting" (all data for each contact is contained in each
HID report packet sent to the host).  The demo can be modified to support additional
contacts if the report descriptor in usb_config.h is modified, and the appropriate
contact data is sent to the host for each HID report.
********************************************************************/


/** INCLUDES *******************************************************/
#include <stdint.h>
#include <stdbool.h>

#include <system.h>

#include <usb/usb.h>
#include <usb/usb_device_hid.h>

/** VARIABLES ******************************************************/
/* Some processors have a limited range of RAM addresses where the USB module
 * is able to access.  The following section is for those devices.  This section
 * assigns the buffers that need to be used by the USB module into those
 * specific areas.
 */
#if defined(FIXED_ADDRESS_MEMORY)
    #if defined(COMPILER_MPLAB_C18)
        #pragma udata DEVICE_HID_DIGITIZER_IN_BUFFER=DEVICE_HID_DIGITIZER_IN_BUFFER_ADDRESS
            static unsigned char hid_report_in[HID_INT_IN_EP_SIZE];
        #pragma udata DEVICE_HID_DIGITIZER_OUT_BUFFER=DEVICE_HID_DIGITIZER_OUT_BUFFER_ADDRESS
            static unsigned char hid_report_out[HID_INT_OUT_EP_SIZE];
        #pragma udata
    #elif defined(__XC8)
        static unsigned char hid_report_in[HID_INT_IN_EP_SIZE] DEVICE_HID_DIGITIZER_IN_BUFFER_ADDRESS;
        static unsigned char hid_report_out[HID_INT_OUT_EP_SIZE] DEVICE_HID_DIGITIZER_OUT_BUFFER_ADDRESS;
    #endif
#else
    static unsigned char hid_report_in[HID_INT_IN_EP_SIZE];
    static unsigned char hid_report_out[HID_INT_OUT_EP_SIZE];
#endif
    
static uint16_t DebounceCounter;
static bool SwitchDebouncing;
static uint8_t TimesPressed;		//Keeps track of how many times the button was pressed, to decide what to do next
static bool FlickPending;		//TRUE if we are busy sending a flick, FALSE if no data is being sent right now
static uint8_t ReportsSent;
static bool CoordinatesChanging;
static uint16_t XCoordinate1;
static uint16_t YCoordinate1;
static uint16_t XCoordinate2;
static uint16_t YCoordinate2;
static uint16_t XCoordinate3;
static uint16_t YCoordinate3;
static uint16_t XOrigin1;
static uint16_t YOrigin1;
static uint16_t XOrigin2;
static uint16_t YOrigin2;
static uint16_t XOrigin3;
static uint16_t YOrigin3;
static uint16_t SOFCount;			//Can get updated in interrupt context if using USB interrupts instead of polling
static uint16_t SOFCountIntSafe;	//Shadow holding variable for the SOFCounter.  This number will not get updated in the interrupt context.
static uint16_t SOFCountIntSafeOld;
static USB_HANDLE lastTransmission;

//Mouse emulation mode variables
static bool EmulateMouseMovement;
static uint8_t movement_length;
static uint8_t vector;
//The direction that the mouse will move in
static const signed char dir_table[]={-4,-4,-4, 0, 4, 4, 4, 0};

static bool HIDApplicationModeChanging;
static uint8_t DeviceIdentifier;
static uint8_t DeviceMode;

/** DEFINITIONS ****************************************************/
//DeviceMode variable values.  See also the usb_config.h "DEFAULT_DEVICE_MODE" definition.
#define MOUSE_MODE                      0x00
#define SINGLE_TOUCH_DIGITIZER_MODE     0x01
#define MULTI_TOUCH_DIGITIZER_MODE      0x02

#define DEBOUNCE_COUNT                  50

/** Private Prototypes *********************************************/
static void Emulate_Digitizer(void);
static void Emulate_Mouse(void);
static void CheckForOverUnderFlow(void);
static void USBHIDCBSetReportComplete(void);
static void DebounceSwitch(void);

/*********************************************************************
* Function: void APP_DeviceHIDDigitizerInitialize(void);
*
* Overview: Initializes the demo code
*
* PreCondition: None
*
* Input: None
*
* Output: None
*
********************************************************************/
void APP_DeviceHIDDigitizerInitialize()
{
    //enable the HID endpoint
    USBEnableEndpoint(HID_EP,USB_IN_ENABLED|USB_HANDSHAKE_ENABLED|USB_DISALLOW_SETUP);

    //Used for sw3 pushbutton debouncing
    TimesPressed = 0;		//Tracks number of times the pushbutton is pressed
    DebounceCounter = 0;	//Used for pushbutton debounce
    SwitchDebouncing = false;

	//Initialize digitizer emulation variables.
    FlickPending = false; 	//Not currently sending a gesture

    //initialize the variable holding the handle for the last
    // transmission
    lastTransmission = 0;

    HIDApplicationModeChanging = false;

    //Initialize device mode and digitizer emulation variables.
    //--------------------------------------------------------
    DeviceIdentifier = 0x01;
    DeviceMode = DEFAULT_DEVICE_MODE;	//Set the device mode (ex: mouse, single-touch digitizer, multi-touch digitizer) to a default value.  Set the default in the usb_config.h file.
    FlickPending = false; 				//Not currently sending a gesture

    //Initialize Mouse mode variables
    //--------------------------------------------------------
    movement_length = 0xFF;
    //State variable for deciding if the mouse data sent to the host should
    //move the cursor or not.
    EmulateMouseMovement = true;
}

/*********************************************************************
* Function: void APP_DeviceHIDDigitizerInitialize(void);
*
* Overview: Initializes the demo code
*
* PreCondition: None
*
* Input: None
*
* Output: None
*
********************************************************************/
void APP_DeviceHIDDigitizerSOFHandler()
{
    // No need to clear UIRbits.SOFIF to 0 here.
    // Callback caller is already doing that.

    //Using SOF packets (which arrive at 1ms intervals) for time
    //keeping purposes in this demo.  This enables this demo to send constant
    //velocity contact point X/Y coordinate movement.
    SOFCount++;
    
    DebounceSwitch();
}


/*********************************************************************
* Function: void APP_DeviceHIDDigitizerTasks(void);
*
* Overview: Keeps the Custom HID demo running.
*
* PreCondition: The demo should have been initialized and started via
*   the APP_DeviceHIDDigitizerInitialize() and APP_DeviceHIDDigitizerStart() demos
*   respectively.
*
* Input: None
*
* Output: None
*
********************************************************************/
void APP_DeviceHIDDigitizerTasks()
{
    /* If the device is not configured yet, or if the device is suspended
     * then exit this function immediately since we can't actually send any
     * application data.
     */
    if( (USBGetDeviceState() < CONFIGURED_STATE) || 
        (USBIsDeviceSuspended() == true) )
    {
        return;
    }
    
    //Don't want to send any report packets on EP1 IN to the host when
    //the host is in the process of sending a control transfer (ex: SET_REPORT)
    //and is changing the device mode.  Need to wait until the control transfer
    //is complete and the new device mode setting has taken effect.
    if(HIDApplicationModeChanging == true)
    {
            return;
    }

    //Check what HID device mode we are currently in, to decide what type of reports
    //should be sent to the host.
    if(DeviceMode == MOUSE_MODE)
    {
        //If the user presses the sw3 pushbutton, start or stop sending mouse
        //movement data
        if((BUTTON_IsPressed(BUTTON_DEVICE_HID_DIGITIZER_DEMO) == true) && (SwitchDebouncing == false))
        {
            SwitchDebouncing = true;
            //If the mouse was moving, stop moving the cursor, if it wasn't
            //moving, start sending data to move mouse cursor in a circle
            EmulateMouseMovement = !EmulateMouseMovement;
        }

        //Prepare the mouse data and send it to the host.
        Emulate_Mouse();

    }//if(DeviceMode == MOUSE_MODE)
    else if((DeviceMode == MULTI_TOUCH_DIGITIZER_MODE) || (DeviceMode == SINGLE_TOUCH_DIGITIZER_MODE))
    {
        //Need to be careful when reading the SOFCounter variable.  If USB interrupts are enabled,
        //the SOFCounter can potentially get updated while you are still reading from it.
        USBMaskInterrupts();
        SOFCountIntSafe = SOFCount;
        USBUnmaskInterrupts();
        //We use the SOFCounterIntSafe variable as a form of clock.  Microsoft's
        //Windows Touch related docoumentation implies that Windows can use
        //"additional heuristics" in determining if data sent from a touch device
        //is valid or not.  If for instance the reported contact point suddenly
        //jumps a great distance (apparent velocity greatly exceeding anything that
        //a real human contact would produce), the operating system may conceivably
        //choose to discard the evidently impossible data.
        //Therefore, it is important that data sent to the host be "realistic"
        //to ensure that it gets processed correctly.  In order to make this
        //easier to accomplish in this demo, we use the SOF packets (sent once every 1ms
        //on a full speed USB bus) as a timer, so that we can use it to vary our reported
        //contact point X and Y coordinates in a smooth real time "believable" (constant
        //velocity) manner.  In a real application with a real digitizer touch pad, it
        //may not necessarily be required to independently keep track of "time", since
        //presumably the input data is being generated in a realisitic real-time fashion
        //by the human interacting with it.  Specific requirements will however vary
        //based on the type of touch contact point sensing technology employed (ex:
        //optical, capacitive, resistive, etc.).

        //Check if the pushbutton is currently pressed.
        //If the user presses the pushbutton, send a flurry of packets emulating a
        //touch (or multi-touch) gesture
        if((BUTTON_IsPressed(BUTTON_DEVICE_HID_DIGITIZER_DEMO) == true) && (SwitchDebouncing == false))
        {
            SwitchDebouncing = true;
            FlickPending = true;
            ReportsSent = 0;
            USBMaskInterrupts();
            SOFCount = 0;
            USBUnmaskInterrupts();
            SOFCountIntSafe = 0;
            SOFCountIntSafeOld = 0;
            CoordinatesChanging = true;
            //Number of times the pushbutton has been pressed by the user.
            //TimesPressed determines what gesture will be sent to the host.
            TimesPressed++;				  
            if(TimesPressed >= 12u)
            {
                    TimesPressed = 1; //Go back to first gesture to repeat the sequence.
            }

            //Emulated thumb finger contact point originating at center of screen
            XOrigin1 = 2400u;	//From report descriptor, valid X-coordinate values are 0-4800
            YOrigin1 = 1500u;	//From report descriptor, valid Y-coordinate values are 0-3000

            //Emulated index finger positioned to upper right of thumb
            XOrigin2 = XOrigin1 + 400u;
            YOrigin2 = YOrigin1 - 400u;

            //Place third contact point somwhere else (in this case upper left of contact 1), 
            //far enough away from contact 1 and contact 2, so that the PC software knows 
            //it is a unique and sensible/valid contact point.  
            XOrigin3 = XOrigin1 - 1000u;
            YOrigin3 = YOrigin1 - 800u;
		
            XCoordinate1 = XOrigin1;
            YCoordinate1 = YOrigin1;
            XCoordinate2 = XOrigin2;
            YCoordinate2 = YOrigin2;
            XCoordinate3 = XOrigin3;
            YCoordinate3 = YOrigin3;
        }

        //Send data for a gesture if needed
        if(FlickPending == true)
        {
            Emulate_Digitizer();
        }

    }//else if(DeviceMode == MULTI_TOUCH_DIGITIZER_MODE)
}

/******************************************************************************
 * Function:        void DebounceSwitch(void)
 *
 * PreCondition:    Switch sw3 should be already initialized.
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:
 *
 * Overview:    This is a non-blocking function that checks the sw3 pin
 *              state.  It uses a counter to make sure the user is no longer
 *              pressing the button, and the value has been steady for a long
 *              time before it "re-enables" the button so that it may
 *              register the next user press action.
 *
 * Note:
 *****************************************************************************/
static void DebounceSwitch(void)
{
    //Perform non-blocking switch debounce operation.
    //User must remove finger from button for awhile before the button becomes
    //active again.
    if(BUTTON_IsPressed(BUTTON_DEVICE_HID_DIGITIZER_DEMO) == true)
    {
        //Immediately reset the counter if the pin is (still) low.
        DebounceCounter = 0;
    }
    else if((BUTTON_IsPressed(BUTTON_DEVICE_HID_DIGITIZER_DEMO) == false) && (SwitchDebouncing == true))
    {
        //Keep track of how long it has been continuously pressed without interruption.
        DebounceCounter++;
    }
    if((SwitchDebouncing == true) && (DebounceCounter == DEBOUNCE_COUNT))
    {
        //Button was continuously released long enough that it must be done bouncing
        DebounceCounter = 0;
        SwitchDebouncing = false;
    }
}

/******************************************************************************
 * Function:        void Emulate_Mouse(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    The ownership of the USB buffers will change according
 *                  to the required operation
 *
 * Overview:    This routine will emulate the function of the mouse.  It
 *              does this by sending IN packets of data to the host.
 *
 *              The generic function HIDTxPacket() is used to send HID
 *              IN packets over USB to the host.
 *
 * Note:        The HID input reports for this mode should be formatted as
 *              follows (based on the report descriptor top level mouse
 *              collection):
 *
 *      //byte[0] = Report ID = MOUSE_DATA_REPORT_ID
 *      //byte[1] = contains bit fields for various input information typically generated by mouse buttons or touch pad tap operations. '1' is the active value (ex: pressed), '0' is the non active value
 *      //		bit0 = Button 1.
 *      //		bit1 = Button 2.
 *      //		bit2 = Button 3.
 *      //		bits3-7 = pad bits, no valid data.
 *      //byte[2] = Relative X coordinate change since the last report was sent
 *      //byte[3] = Relative Y coordinate change since the last report was sent
 *
 *****************************************************************************/
static void Emulate_Mouse(void)
{
    //This function sends packets to the host that make the mouse cursor
    //move in a semi-circular pattern.  In a real application, which has been
    //designed to support multiple device modes (to include mouse mode), the
    //input packet data would typically be generated from the touch-screen
    //user interface.  Since a mouse does not report absolute position
    //information to the host (it only sends relative change in X and Y
    //coordinates since the last packet, instead of absolute X and Y
    //coordinates), this firmware should be replaced with code that uses the
    //touch interface to function like a touch pad.

    //Make sure that the endpoint is not busy before modifying the hid_report_in buffer.
    if(!USBHandleBusy(lastTransmission))
    {
        //Prepare the HID input report containing the mouse movement data.
        hid_report_in[0] = MOUSE_DATA_REPORT_ID;		//This is the report ID byte.  Mouse report ID = 0x04 from report descriptor.

        if(EmulateMouseMovement == true)
        {
            //go 14 times in the same direction before changing direction
            if(movement_length > 14u)
            {
                hid_report_in[1] = 0u;								//Button click data
                hid_report_in[2] = dir_table[vector & 0x07];           // X-Vector
                hid_report_in[3] = dir_table[(vector+2) & 0x07];       // Y-Vector
                //go to the next direction in the table
                vector++;
                //reset the counter for when to change again
                movement_length = 0;
            }//end if(movement_length > 14u)
        }
        else
        {
            //don't move the mouse
            hid_report_in[1] = hid_report_in[2] = hid_report_in[3] = 0;
        }

        //Send the 4 byte packet over USB to the host.
        lastTransmission = HIDTxPacket(HID_EP, (uint8_t*)hid_report_in, 0x04);

        //increment the counter of when to change the data sent
        movement_length++;

    }//if(!USBHandleBusy(lastTransmission))
}//end Emulate_Mouse


/******************************************************************************
 * Function:     	void Emulate_Digitizer(void)
 *
 * PreCondition: 	None
 *
 * Input:        	None
 *
 * Output:       	None
 *
 * Side Effects:
 *          The ownership of the USB buffers will change according
 *          to the required operation
 *
 * Overview:
 *          This routine will emulate the function of a human interacting
 *          with a touch digitizer panel.  It does this by sending IN packets
 *          of data to the host, containing data simulating that which would
 *          be generated by real human contact(s).
 *
 *          The generic function HIDTxPacket() is used to send HID
 *          IN packets over USB to the host.
 *
 *          This function is used to emulate the digitizer behavior, in
 *          both the single-touch and multi-touch device modes (DeviceMode variable).
 *          The formatting and content of the data being sent to the host depends
 *          on the current device mode.  On OSes that support multi-touch
 *          capability, the host will configure the DeviceMode to
 *          the MULTI_TOUCH_DIGITIZER_MODE. By default however, it is best
 *          if the device powers up with some other default mode (mouse mode
 *          or single-touch digizer), so as to provide improved functionality
 *          on legacy operating systems (that don't support multi-touch
 *          digitizers, and would not know how to interpret/use the reports).
 *
 *
 * Note:    Reports of data should be sent to the host matching the Report
 *          Descriptor and the current DeviceMode.
 *          The HID report (IN) packets sent to the host should be formatted like the
 *          report descriptor in the usb_descriptors.c file.  See the usb_descriptors.c
 *          file for additional comments.  For this demo, we use the following report
 *          format:
 *
 *	//Report format for use in multi-touch device mode, which can support up to two
 *	//simultaneous contacts, using the report descriptor provided in usb_descriptors.c:
 *
 *	//Byte[0] = Report ID == use literal value = MULTI_TOUCH_DATA_REPORT_ID when sending contact reports.
 * 	//First contact point info in bytes 1-6.
 *	//Byte[1] = Bits7-2: pad bits (unused), Bit1:In Range, Bit0:Tip Switch
 *	//Byte[2] = Contact identifier number (assigned by firmware, see report descriptor)
 *	//Byte[3] = X-coordinate LSB
 *	//Byte[4] = X-coordinate MSB
 *	//Byte[5] = Y-coordinate LSB
 *	//Byte[6] = Y-coordinate MSB
 *
 *	//Second contact point info in bytes 7-12
 *	//Byte[7] = Bits7-2: pad bits (unused), Bit1:In Range, Bit0:Tip Switch
 *	//Byte[8] = Contact identifier number (assigned by firmware, see report descriptor)
 *	//Byte[9] = X-coordinate LSB
 *	//Byte[10]= X-coordinate MSB
 *	//Byte[11]= Y-coordinate LSB
 *	//Byte[12]= Y-coordinate MSB
 *
 *	//Third contact point info in bytes 13-18
 *	//Byte[13] = Bits7-2: pad bits (unused), Bit1:In Range, Bit0:Tip Switch
 *	//Byte[14] = Contact identifier number (assigned by firmware, see report descriptor)
 *	//Byte[15] = X-coordinate LSB
 *	//Byte[16] = X-coordinate MSB
 *	//Byte[17] = Y-coordinate LSB
 *	//Byte[18] = Y-coordinate MSB
 *
 *	//Contacts valid
 *	//Byte[19]= 8-bit number indicating how many of the above contact points are valid.  
 *				If only the first contact is valid, report "1" here.  If 2 are 
 *				valid, report "2".  If three are valid, report "3".
 *
 *	When operating in the DeviceMode == SINGLE_TOUCH_DIGITIZER_MODE
 *	The input data reports are different.  In this mode, the device should only keep
 *	track of and send information to the host for the first contact point.  In this
 *	mode, the report sent to the host should be formatted as:
 *	//byte[0] = Report ID = SINGLE_TOUCH_DATA_REPORT_ID
 *	//byte[1] = contains bit fields for various input information typically generated by an input pen or human finger. '1' is the active value (ex: pressed), '0' is the non active value
 *	//		bit0 = Tip switch. At the end of a pen input device would normally be a pressure senstive switch.  Asserting this performs an operation analogous to a "left click" on a mouse
 *	//		bit1 = In range indicator.  
 *	//		bit2 though bit7 = Pad bits.  Values not used for anything.
 *	//byte[2] = Pad byte.  Value not used for anything.
 *	//byte[3] = X coordinate LSB value of contact point
 *	//byte[4] = X coordinate MSB value of contact point
 *	//byte[5] = Y coordinate LSB value of contact point
 *	//byte[6] = Y coordinate MSB value of contact point
 *
 *****************************************************************************/
static void Emulate_Digitizer(void)
{
    //Make sure the endpoint buffer (in this case hid_report_in[] is not busy
    //before we modify the contents of the buffer for the next transmission.
    if(!USBHandleBusy(lastTransmission))
    {
        //Find the time delay in milliseconds (number of SOF packets) since
        //the last transmission.  We will use this delta number to establish
        //how much our X/Y coordinates should change (for simulated constant
        //velocity movement).
        uint16_t SOFDelta = (SOFCountIntSafe - SOFCountIntSafeOld);

        //Report ID for multi-touch contact information reports (based on report descriptor)
        hid_report_in[0] = MULTI_TOUCH_DATA_REPORT_ID;		//Report ID in byte[0]

        hid_report_in[1] = 0b00000010;	//Contact1: In-range, but tip switch not pressed.
        hid_report_in[7] = 0b00000010;	//Contact2: In-range, but tip switch not pressed.
        hid_report_in[13]= 0b00000010;	//Contact3: In-range, but tip switch not pressed.
        hid_report_in[19] = 2;			//Assume two contacts valid. (unless we set hid_report_in[19] = 1 or 3 later on in this function)

        //Decide how (or if) we want to change our X and Y Coordinates.
        if(CoordinatesChanging == true)
        {
            hid_report_in[1] = 0b00000011;	//Contact1: Tip switch and in range indicators set.
            hid_report_in[7] = 0b00000011;	//Contact2: Tip switch and in range indicators set.
            hid_report_in[13]= 0b00000011;	//Contact3: Tip switch and in range indicators set.

            //Decide what type of data to send to host, based on how many times the
            //user pressed the pushbutton.
            switch(TimesPressed)
            {
                case 1:	//Emulate a single touch "back" flick operation.
                    //This is done by pressing only one finger down, and by moving it right quickly (X plus)
                    SOFDelta = SOFDelta * 8;	//To get faster movement, we scale our time delay.
                    XCoordinate1 = XCoordinate1 + SOFDelta;
                    if(XCoordinate1 > (XOrigin1 + 246u))
                        CoordinatesChanging = false;
                    hid_report_in[19] = 1;		//Only the first contact point is valid	(this is a single touch gesture)
                    break;

                case 2:	//Emulate a single touch "forward" flick operation.
                    //This is done by pressing only one finger down, and by moving it left quickly (X minus)
                    SOFDelta = SOFDelta * 8;	//To get faster movement, we scale our time delay.
                    XCoordinate1 = XCoordinate1 - SOFDelta;
                    if(XCoordinate1 < (XOrigin1 - 246u))
                        CoordinatesChanging = false;
                    hid_report_in[19] = 1;		//Only the first contact point is valid	(this is a single touch gesture)
                    break;

                case 3:	//Emulate a multi-touch zoom-in operation ("un-pinch" index and thumb fingers)
                    //Sweep Contact1 (emulated thumb finger) coordinates to move lower left (X minus, Y plus)
                    SOFDelta = SOFDelta >> 1;	//To get slower movement, we scale our time delay.
                    XCoordinate1 = XCoordinate1 - SOFDelta;
                    if(XCoordinate1 < (XOrigin1 - 352u))
                        CoordinatesChanging = false;
                    YCoordinate1 = YCoordinate1 + SOFDelta;
                    if(YCoordinate1 > (YOrigin1 + 300u))
                        CoordinatesChanging = false;
                    //Sweep Contact2 (emulated index finger) coordinates to move upper right (X plus, Y minus)
                    XCoordinate2 = XCoordinate2 + SOFDelta;
                    if(XCoordinate2 > (XOrigin2 + 352u))
                        CoordinatesChanging = false;
                    YCoordinate2 = YCoordinate2 - SOFDelta;
                    if(YCoordinate2 < (YOrigin2 - 300u))
                        CoordinatesChanging = false;
                    break;

                case 4:	//Emulate a multi-touch zoom-in operation ("un-pinch" index and thumb fingers)
                    //Sweep Contact1 (emulated thumb finger) coordinates to move lower left (X minus, Y plus)
                    SOFDelta = SOFDelta >> 1;	//To get slower movement, we scale our time delay.
                    XCoordinate1 = XCoordinate1 - SOFDelta;
                    if(XCoordinate1 < (XOrigin1 - 352u))
                        CoordinatesChanging = false;
                    YCoordinate1 = YCoordinate1 + SOFDelta;
                    if(YCoordinate1 > (YOrigin1 + 300u))
                        CoordinatesChanging = false;
                    //Sweep Contact2 (emulated index finger) coordinates to move upper right (X plus, Y minus)
                    XCoordinate2 = XCoordinate2 + SOFDelta;
                    if(XCoordinate2 > (XOrigin2 + 352u))
                        CoordinatesChanging = false;
                    YCoordinate2 = YCoordinate2 - SOFDelta;
                    if(YCoordinate2 < (YOrigin2 - 300u))
                        CoordinatesChanging = false;
                    break;

                case 5:	//Emulate a multi-touch zoom-in operation ("un-pinch" index and thumb fingers)
                    //Sweep Contact1 (emulated thumb finger) coordinates to move lower left (X minus, Y plus)
                    SOFDelta = SOFDelta >> 1;	//To get slower movement, we scale our time delay.
                    XCoordinate1 = XCoordinate1 - SOFDelta;
                    if(XCoordinate1 < (XOrigin1 - 352u))
                        CoordinatesChanging = false;
                    YCoordinate1 = YCoordinate1 + SOFDelta;
                    if(YCoordinate1 > (YOrigin1 + 300u))
                        CoordinatesChanging = false;
                    //Sweep Contact2 (emulated index finger) coordinates to move upper right (X plus, Y minus)
                    XCoordinate2 = XCoordinate2 + SOFDelta;
                    if(XCoordinate2 > (XOrigin2 + 352u))
                        CoordinatesChanging = false;
                    YCoordinate2 = YCoordinate2 - SOFDelta;
                    if(YCoordinate2 < (YOrigin2 - 300u))
                        CoordinatesChanging = false;
                    break;

                case 6:	//Emulate a two-finger horizontal scroll right operation
                    //Tap both fingers down simultaneously and move them both to the left (X minus)
                    XCoordinate1 = XCoordinate1 - SOFDelta;
                    if(XCoordinate1 < (XOrigin1 - 586u))
                        CoordinatesChanging = false;
                    XCoordinate2 = XCoordinate2 - SOFDelta;
                    if(XCoordinate2 < (XOrigin2 - 586u))
                        CoordinatesChanging = false;
                    break;

                case 7:	//Emulate a two-finger horizontal scroll left operation
                    //Tap both fingers down simultaneously and move them both to the right (X plus)
                    XCoordinate1 = XCoordinate1 + SOFDelta;
                    if(XCoordinate1 > (XOrigin1 + 586u))
                        CoordinatesChanging = false;
                    XCoordinate2 = XCoordinate2 + SOFDelta;
                    if(XCoordinate2 > (XOrigin2 + 586u))
                        CoordinatesChanging = false;
                    break;

                case 8:	//Emulate a single touch scroll down operation.
                    //This is done by pressing only one finger down, and by moving it down (Y plus)
                    YCoordinate1 = YCoordinate1 + SOFDelta;
                    if(YCoordinate1 > (YOrigin1 + 500u))
                        CoordinatesChanging = false;
                    hid_report_in[13] = 1;		//Only the first contact point is valid
                    break;

                case 9:	//Emulate a single touch scroll up operation.
                    //This is done by pressing only one finger down, and by moving it up (Y minus)
                    YCoordinate1 = YCoordinate1 - SOFDelta;
                    if(YCoordinate1 < (YOrigin1 - 500u))
                        CoordinatesChanging = false;
                    hid_report_in[19] = 1;		//Only the first contact point is valid
                    break;

                case 10://Emulate a multi-touch zoom-out operation (pinch two fingers together)
                    //Sweep Contact1 (emulated thumb) coordinates to move upper right (X plus, Y minus)
                    SOFDelta = SOFDelta >> 1;	//To get slower movement, we scale our time delay.
                    XCoordinate1 = XCoordinate1 + SOFDelta;
                    if(XCoordinate1 > (XOrigin1 + 352u))
                        CoordinatesChanging = false;
                    YCoordinate1 = YCoordinate1 - SOFDelta;
                    if(YCoordinate1 < (YOrigin1 - 300u))
                        CoordinatesChanging = false;
                    //Sweep Contact2 (emulated index finger) coordinates to move lower left (X minus, Y plus)
                    XCoordinate2 = XCoordinate2 - SOFDelta;
                    if(XCoordinate2 < (XOrigin2 - 352u))
                        CoordinatesChanging = false;
                    YCoordinate2 = YCoordinate2 + SOFDelta;
                    if(YCoordinate2 > (YOrigin2 + 300u))
                        CoordinatesChanging = false;
                    break;
					
                case 11:    //Three contact point valid example case.  Since Windows 7 built in gestures only need/use 2 contact points,
                    //some other PC application (which is capable of using the WM_GESTURE and other windows messages) to receive
                    //all three contact points will be necessary to demonstrate this functionality.
                    SOFDelta = SOFDelta >> 1;	//To get slower movement, we scale our time delay.
                    //For this example case, lets sweep contact 1 from left to right (+X direction).  For contact 2,
                    //lets sweep contact in the -Y direction (towards the top of the screen).  For contact 3,
                    //lets sweep the contact point in both the +X and -Y directions (diagonally up/right of screen).
                    XCoordinate1 = XCoordinate1 + SOFDelta;
                    YCoordinate2 = YCoordinate2 - SOFDelta;
                    XCoordinate3 = XCoordinate3 + SOFDelta;
                    YCoordinate3 = YCoordinate3 - SOFDelta;
                    hid_report_in[19] = 3;      //Three contacts valid.
                    break;
                default:
                    break;
            } //end switch(TimesPressed)
        } //end if(CoordinatesChanging == true)

        //Make sure none of the above math operations inadvertently exceeded the
        //dimensions of our touch panel.
        CheckForOverUnderFlow();

        //First contact info in bytes 1-6
        hid_report_in[2] = 1;						//Contact ID
        hid_report_in[3] = XCoordinate1;			//X-coord LSB
        hid_report_in[4] = (XCoordinate1 >> 8);		//X-coord MSB
        hid_report_in[5] = YCoordinate1;			//Y-coord LSB
        hid_report_in[6] = (YCoordinate1 >> 8);		//Y-coord MSB

        //Second contact info in bytes 7-12
        //hid_report_in[] bytes 0, 1, 7, and 13 were initialized earlier in this function
        if(DeviceMode == MULTI_TOUCH_DIGITIZER_MODE)
        {
            //Second contact point in bytes 7-12
            hid_report_in[8] = 2;						//Contact ID
            hid_report_in[9] = XCoordinate2;			//X-coord LSB
            hid_report_in[10] = (XCoordinate2 >> 8);	//X-coord MSB
            hid_report_in[11] = YCoordinate2;			//Y-coord LSB
            hid_report_in[12] = (YCoordinate2 >> 8);	//Y-coord MSB

            //Third contact point in bytes 13-18
            hid_report_in[14] = 3;						//Contact ID
            hid_report_in[15] = XCoordinate3;			//X-coord LSB
            hid_report_in[16] = (XCoordinate3 >> 8);	//X-coord MSB
            hid_report_in[17] = YCoordinate3;			//Y-coord LSB
            hid_report_in[18] = (YCoordinate3 >> 8);	//Y-coord MSB

            //Arm the endpoint so that the packet can be sent over USB to the host.
            //The HIDTxPacket() function initializes the BDT entry for the specified endpoint.
            //Always make sure that the USBHandleBusy() returned false before calling
            //the HIDTxPacket() function (this check was done at the top of Emulate_Digitizer()).
            lastTransmission = HIDTxPacket(HID_EP, (uint8_t*)hid_report_in, 20);
        }
        else //must be DeviceMode == SINGLE_TOUCH_DIGITIZER_MODE
        {
            hid_report_in[0] = SINGLE_TOUCH_DATA_REPORT_ID;		//Report ID in byte[0].
            //The hid_report_in[2] byte is the contact ID in MULTI_TOUCH_DIGITIZER_MODE.
            //In SINGLE_TOUCH_DIGITIZER_MODE, this is a pad byte with no valid data.
            hid_report_in[2] = 0x00;

            //Send the packet now.
            lastTransmission = HIDTxPacket(HID_EP, (uint8_t*)hid_report_in, 7);
        }


       	//Counter used to keep track of how many packets have been sent, to decide what to do next
       	ReportsSent++;

        if(ReportsSent == 70)
        {
            //but we still want to send a few more packets with the tip switch off, and no movement
            CoordinatesChanging = false;
        }
        if(ReportsSent == 80)
        {
            //We finished sending the flick + "idle".  No need to send more packets for now.
            FlickPending = false;
        }

        //Save the SOF (time) value so we can use it to know the elapsed time
        //until the next instance that Emulate_Digitizer() gets called.
        SOFCountIntSafeOld = SOFCountIntSafe;

    }//end if(HIDTxHandleBusy(lastTransmission) == 0)

}//end Emulate_Digitizer



/********************************************************************
 * Function:        void UserGetReportHandler(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        UserGetReportHandler() is used to respond to
 *                      the HID device class specific GET_REPORT control
 *                      transfer request (starts with SETUP packet on EP0 OUT).
 *                  This get report handler callback function is placed
 *                      here instead of in usb_function_hid.c, since this code
 *                      is "custom" and needs to be specific to this particular
 *                      application example.  For this HID digitizer device,
 *                      we need to be able to correctly respond to
 *                      GET_REPORT(feature) requests.  The proper response to this
 *                      type of request depends on the HID report descriptor (in
 *                      usb_descriptors.c).
 * Note:
 *******************************************************************/
void UserGetReportHandler(void)
{
    static uint8_t FeatureReport[3];
    uint8_t bytesToSend;

    //Check if request was for a feature report with report ID = VALID_CONTACTS_FEATURE_REPORT_ID = 0x02.
    //wValue MSB = 0x03 is Feature report (see HID1_11.pdf specifications),
    //LSB = VALID_CONTACTS_FEATURE_REPORT_ID is Report ID "0x02".  This feature report ID
    //is used for reporting the maximum number of supported simultaneous contacts (in multi-touch mode).
    if(SetupPkt.wValue == (0x0300 + VALID_CONTACTS_FEATURE_REPORT_ID))
    {
        static uint8_t FeatureReport[2];

        //Prepare a response packet for the host
        FeatureReport[0] = VALID_CONTACTS_FEATURE_REPORT_ID;
        FeatureReport[1] = MAX_VALID_CONTACT_POINTS;	//Three contacts valid for this multi-touch demo (can be increased by increasing this number and editing report descriptor and report data payload)

        //Determine number of bytes to send to the host
        if(SetupPkt.wLength < 2u)
        {
            bytesToSend = SetupPkt.wLength;
        }
        else
        {
            //Size of the feature report.  Byte 0 is Report ID = 0x02, Byte 1 is
            //the maximum contacts value, which is "2" for this demo
            bytesToSend = 2;
        }

        //Now send the reponse packet data to the host, via the control transfer on EP0
        USBEP0SendRAMPtr((uint8_t*)&FeatureReport, bytesToSend, USB_EP0_RAM);
    }
    else if(SetupPkt.wValue == (0x0300 + DEVICE_MODE_FEATURE_REPORT_ID))	//Report ID = 0x03 = Configuration report, which is used to select between mouse, single/multi digitizer modes.
    {
        //Prepare a response packet for the host
        FeatureReport[0] = DEVICE_MODE_FEATURE_REPORT_ID;
        FeatureReport[1] = DeviceMode;
        FeatureReport[2] = DeviceIdentifier;

        //Determine number of bytes to send to the host
        if(SetupPkt.wLength < 3u)
        {
            bytesToSend = SetupPkt.wLength;
        }
        else
        {
            //Size of the feature report.  Byte 0 is report ID, byte 1 is the
            //current DeviceMode, Byte 2 is the DeviceIdenfier
            bytesToSend = 3;
        }

        //Now send the reponse packet data to the host, via the control transfer on EP0
        USBEP0SendRAMPtr((uint8_t*)&FeatureReport, bytesToSend, USB_EP0_RAM);
    }
}

/********************************************************************
 * Function:        void UserSetReportHandler(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:    UserSetReportHandler() is used to respond to
 *              the HID device class specific SET_REPORT control
 *              transfer request (starts with SETUP packet on EP0 OUT).
 *
 *              This get report handler callback function is placed
 *              here instead of in usb_function_hid.c, since this code
 *              is "custom" and needs to be specific to this particular
 *              application example.  For this HID digitizer device, which
 *              supports multiple device modes, we need to be able to
 *              correctly respond to SET_FEATURE requests.  The proper response
 *              to this type of request depends on the HID report descriptor (in
 *              usb_descriptors.c).
 * Note:
 *******************************************************************/
void UserSetReportHandler(void)
{
    if(SetupPkt.wValue == (0x0300 + DEVICE_MODE_FEATURE_REPORT_ID))	//Host is setting the device mode (ex: mouse, single-touch digitizer, multi-touch digitizer)
    {
        //Temporarily stop sending HID report data packets on EP1 IN until
        //the new mode has been fully set/takes effect.
        HIDApplicationModeChanging = true;

        //If firmware gets to here, the host PC is trying to change the device mode.
        //If the EP1 IN endpoint was already armed and filled with report data (from
        //the previous mode setting, ex: a mouse report), then the first packet
        //sent to the host after the SET_REPORT control transfer completes might
        //contain the wrong data/be for the wrong mode.  To ensure that all packets
        //are consistent to the currently configured mode, even during switchover
        //events, special handling should be done to dis-arm EP1 IN until the new
        //mode setting is complete.

        //Check if the EP1 IN endpoint was already armed.  If so, need to abandon the
        //operation to avoid sending any packets, until the new mode has been set.
        if(USBHandleBusy(lastTransmission))	//effectively: if(EP1 IN UOWN bit == 1)
        {
            //In this case we need to terminate the pending USB transfers on the
            //  HID endpoint.
            USBCancelIO(HID_EP);
        }

        //Prepare EP0 to receive the control transfer data (the device mode to set)
        USBEP0Receive((uint8_t*)&hid_report_out, SetupPkt.wLength, USBHIDCBSetReportComplete);	//Host will send two bytes.  After the two bytes are successfully received, call the USBHIDCBSetReportComplete() callback function.
    }
}


//Secondary callback function that gets called when the above
//control transfer,  USBEP0Receive(...,...,USBHIDCBSetReportComplete) completes.
static void USBHIDCBSetReportComplete(void)
{
    //The hid_report_out[0] byte contains the Report ID that is getting set.
    //The hid_report_out[1] byte contains the Device Mode
    //The hid_report_out[2] byte contains the DeviceIdentifier
    
    if(hid_report_out[0] == DEVICE_MODE_FEATURE_REPORT_ID)	//configuration setting feature Report ID = 0x03 = DEVICE_MODE_FEATURE_REPORT_ID
    {
        //Set the device variables now, so the firmware can behave according to the new
        //settings that have been specified by the host.

        //Error checking.  Make sure the host is trying to choose a valid setting.
        if((hid_report_out[1] == MOUSE_MODE) || (hid_report_out[1] == SINGLE_TOUCH_DIGITIZER_MODE) || (hid_report_out[1] == MULTI_TOUCH_DIGITIZER_MODE))
        {
            DeviceMode = hid_report_out[1];		//The DeviceMode variable is used else where in the application firmware to determine the behavior/type of input reports sent to the host.
            DeviceIdentifier = hid_report_out[2];
        }
    }

    //The new device mode setting has been set.  Okay to start sending HID report
    //packets again on EP1 IN now.
    HIDApplicationModeChanging = false;
}


static void CheckForOverUnderFlow(void)
{
    //Code to check for overflow and underflow conditions, to make sure we aren't
    //inadvertently trying to report out of range X or Y coordinates.

    //If underflow occurred, the MSb (negative) bit will be set. The coordinates
    //are stored as WORDs (16-bit unsigned integers)
    if(XCoordinate1 & 0x8000)
        XCoordinate1 = 0x0000;
    if(YCoordinate1 & 0x8000)
        YCoordinate1 = 0x0000;
    if(XCoordinate2 & 0x8000)
        XCoordinate2 = 0x0000;
    if(YCoordinate2 & 0x8000)
        YCoordinate2 = 0x0000;
    if(XCoordinate3 & 0x8000)
        XCoordinate3 = 0x0000;
    if(YCoordinate3 & 0x8000)
        YCoordinate3 = 0x0000;

    //Check for overflow conditions.  If overflow occurred, the coordinate will
    //be higher than the allowed values indicated in the report descriptor.
    if(XCoordinate1 > 4800u)
        XCoordinate1 = 4800u;
    if(YCoordinate1 > 3000u)
        YCoordinate1 = 3000u;
    if(XCoordinate2 > 4800u)
        XCoordinate2 = 4800u;
    if(YCoordinate2 > 3000u)
        YCoordinate2 = 3000u;
    if(XCoordinate3 > 4800u)
        XCoordinate3 = 4800u;
    if(YCoordinate3 > 3000u)
        YCoordinate3 = 3000u;
}