/********************************************************************
 *******************************************************************/

#include <system.h>
#include <system_config.h>
#include <usb/usb.h>
#include "fileio/fileio.h"
#include "driver/fileio/sd_spi.h"

/** CONFIGURATION Bits **********************************************/
#pragma config FWDTEN = OFF
#pragma config ICS = PGx2
#pragma config GWRP = OFF
#pragma config GCP = OFF
#pragma config JTAGEN = OFF

#pragma config POSCMOD = HS
#pragma config IOL1WAY = OFF
#pragma config OSCIOFNC = ON
#pragma config FCKSM = CSDCMD
#pragma config FNOSC = PRIPLL
#pragma config PLL96MHZ = ON
#pragma config PLLDIV = DIV2
#pragma config IESO = OFF

extern FILEIO_SD_DRIVE_CONFIG sdCardMediaParameters;

/*********************************************************************
* Function: void SYSTEM_Initialize(void)
*
* Overview: Initializes the system.
*
* PreCondition: None
*
* Input:  SYSTEM_STATE - the state to initialize the system into
*
* Output: None
*
********************************************************************/
void SYSTEM_Initialize(void)
{
    ANSBbits.ANSB0 = 0;
    ANSBbits.ANSB1 = 0;
    ANSGbits.ANSG6 = 0;
    ANSCbits.ANSC13 = 0;

//	The USB specifications require that USB peripheral devices must never source
//	current onto the Vbus pin.  Additionally, USB peripherals should not source
//	current on D+ or D- when the host/hub is not actively powering the Vbus line.
//	When designing a self powered (as opposed to bus powered) USB peripheral
//	device, the firmware should make sure not to turn on the USB module and D+
//	or D- pull up resistor unless Vbus is actively powered.  Therefore, the
//	firmware needs some means to detect when Vbus is being powered by the host.
//	A 5V tolerant I/O pin can be connected to Vbus (through a resistor), and
// 	can be used to detect when Vbus is high (host actively powering), or low
//	(host is shut down or otherwise not supplying power).  The USB firmware
// 	can then periodically poll this I/O pin to know when it is okay to turn on
//	the USB module/D+/D- pull up resistor.  When designing a purely bus powered
//	peripheral device, it is not possible to source current on D+ or D- when the
//	host is not actively providing power on Vbus. Therefore, implementing this
//	bus sense feature is optional.  This firmware can be made to use this bus
//	sense feature by making sure "USE_USB_BUS_SENSE_IO" has been defined in the
//	HardwareProfile.h file.
    #if defined(USE_USB_BUS_SENSE_IO)
    tris_usb_bus_sense = INPUT_PIN;
    #endif

//	If the host PC sends a GetStatus (device) request, the firmware must respond
//	and let the host know if the USB peripheral device is currently bus powered
//	or self powered.  See chapter 9 in the official USB specifications for details
//	regarding this request.  If the peripheral device is capable of being both
//	self and bus powered, it should not return a hard coded value for this request.
//	Instead, firmware should check if it is currently self or bus powered, and
//	respond accordingly.  If the hardware has been configured like demonstrated
//	on the PICDEM FS USB Demo Board, an I/O pin can be polled to determine the
//	currently selected power source.  On the PICDEM FS USB Demo Board, "RA2"
//	is used for	this purpose.  If using this feature, make sure "USE_SELF_POWER_SENSE_IO"
//	has been defined in HardwareProfile.h, and that an appropriate I/O pin has been mapped
//	to it in HardwareProfile.h.
    #if defined(USE_SELF_POWER_SENSE_IO)
    tris_self_power = INPUT_PIN;
    #endif

    // Configure SPI1 PPS pins
    __builtin_write_OSCCONL(OSCCON & 0xbf); // unlock PPS

    RPOR1bits.RP2R = 8;       // assign RP2 for SCK1
    RPOR0bits.RP1R = 7;       // assign RP1 for SDO1
    RPINR20bits.SDI1R = 0;    // assign RP0 for SDI1

    __builtin_write_OSCCONL(OSCCON | 0x40); // lock   PPS

    FILEIO_SD_IOInitialize(&sdCardMediaParameters);

    USBDeviceInit();	//usb_device.c.  Initializes USB module SFRs and firmware
    					//variables to known states.
}

void USER_SdSpiConfigurePins_1 (void)
{
    //enable a pull-up for the card detect, just in case the SD-Card isn't attached
    //  then lets have a pull-up to make sure we don't think it is there.
    CNPU4bits.CN55PUE = 1;

    // Deassert the chip select pin
    LATGbits.LATG6 = 1;
    // Configure CS pin as an output
    TRISGbits.TRISG6 = 0;
    // Configure CD pin as an input
    TRISAbits.TRISA15 = 1;
    // Configure WP pin as an input
    TRISCbits.TRISC13 = 1;
}

inline void USER_SdSpiSetCs_1(uint8_t a)
{
    LATGbits.LATG6 = a;
}

inline bool USER_SdSpiGetCd_1(void)
{
    return (!PORTAbits.RA15) ? true : false;
}

inline bool USER_SdSpiGetWp_1(void)
{
    return (PORTCbits.RC13) ? true : false;
}


#if defined(USB_INTERRUPT)
void __attribute__((interrupt,auto_psv)) _USB1Interrupt()
{
    USBDeviceTasks();
}
#endif


